///////////////////////////////////////////////////////////////////////////////
// enum.hpp: top level header for BOOST_ENUM
//
// Copyright 2005 Frank Laub
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_ENUM_HPP
#define BOOST_ENUM_HPP

// MS compatible compilers support #pragma once
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
# pragma once
#endif

#include <boost/enum/iterator.hpp>
#include <boost/enum/base.hpp>
#include <boost/enum/bitfield.hpp>
#include <boost/enum/macros.hpp>

#endif
