//
// $Id$
//
//
// Original author: Jarrett Egertson <jegertso .@. uw.edu>
//
// Licensed under the Apache License, Version 2.0 (the "License"); 
// you may not use this file except in compliance with the License. 
// You may obtain a copy of the License at 
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software 
// distributed under the License is distributed on an "AS IS" BASIS, 
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
// See the License for the specific language governing permissions and 
// limitations under the License.
//

#ifndef _DEMUXDATAPROCESSINGSTRINGS_HPP
#define _DEMUXDATAPROCESSINGSTRINGS_HPP

#include <pwiz/utility/misc/Std.hpp>

namespace DemuxDataProcessingStrings
{
    static const std::string kDEMUX_NAME = "Demultiplexing"; ///< This is a flag used by SpectrumWorkerThreads to handle demultiplexing
} // namespace DemuxTypes
#endif //_DEMUXDATAPROCESSINGSTRINGS_HPP
