//
// $Id$
//
//
// Original author: Witold Wolski <wewolski@gmail.com>
//
// Copyright : ETH Zurich
//
// Licensed under the Apache License, Version 2.0 (the "License"); 
// you may not use this file except in compliance with the License. 
// You may obtain a copy of the License at 
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software 
// distributed under the License is distributed on an "AS IS" BASIS, 
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
// See the License for the specific language governing permissions and 
// limitations under the License.
//


#ifndef CONSTANTS_H
#define CONSTANTS_H

#include <string>

/*! \file Constants.h
\brief Constants

*/

namespace ralab
{
  namespace constants
  {
      /// the ratio of the circumference of a circle to its diameter;
      const double PI(3.14159265358979323846264338327950288);
  }//end algo
}//end ralab


#endif // CONSTANTS_H
